import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.py_agata import Agata


def test_analyze_glucose_profile():
    """
    Unit test of Agata.analyze_glucose_profile function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 55, 0), timedelta(minutes=5)).astype(
        datetime)
    glucose = np.zeros(shape=(t.shape[0],))
    glucose[0] = 40
    glucose[1:3] = [60, 60]
    glucose[3] = 80
    glucose[4:6] = [120, 150]
    glucose[6:8] = [200, 200]
    glucose[8:10] = [260, 260]
    glucose[10] = np.nan
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    # Tests

    # Diabetes
    agata = Agata(glycemic_target='diabetes')

    results = agata.analyze_glucose_profile(data=data)
    assert type(results) is dict

    # Variability metrics
    assert type(results) is dict
    assert type(results['variability']) is dict
    assert np.round(results['variability']['mean_glucose'] * 1000) / 1000 == 143.000
    assert np.round(results['variability']['median_glucose'] * 1000) / 1000 == 135.000
    assert np.round(results['variability']['std_glucose'] * 1000) / 1000 == 83.540
    assert np.round(results['variability']['cv_glucose'] * 1000) / 1000 == 58.419
    assert np.round(results['variability']['range_glucose'] * 1000) / 1000 == 220.000
    assert np.round(results['variability']['iqr_glucose'] * 1000) / 1000 == 135.000
    assert np.round(results['variability']['auc_glucose'] * 1000) / 1000 == 7150.000
    assert np.round(results['variability']['gmi'] * 1000) / 1000 == 6.731
    assert np.round(results['variability']['cogi'] * 1000) / 1000 == 19.091
    assert np.isnan(np.round(results['variability']['conga'] * 1000) / 1000)
    assert np.round(results['variability']['j_index'] * 1000) / 1000 == 51.320
    assert np.round(results['variability']['mage_plus_index'] * 1000) / 1000 == 0.0
    assert np.round(results['variability']['mage_minus_index'] * 1000) / 1000 == 0.0
    assert np.round(results['variability']['mage_index'] * 1000) / 1000 == 0.0
    assert np.round(results['variability']['ef_index'] * 1000) / 1000 == 0.0
    assert np.isnan(np.round(results['variability']['modd'] * 1000) / 1000)
    assert np.isnan(np.round(results['variability']['sddm_index'] * 1000) / 1000)
    assert np.round(results['variability']['sdw_index'] * 1000) / 1000 == 83.540
    assert np.round(results['variability']['std_glucose_roc'] * 1000) / 1000 == 1.925
    assert np.round(results['variability']['cvga'] * 1000) / 1000 == 4759.436

    # Time fields
    assert type(results['time_in_ranges']) is dict
    assert results['time_in_ranges']['time_in_target'] == 30
    assert results['time_in_ranges']['time_in_tight_target'] == 20
    assert results['time_in_ranges']['time_in_hypoglycemia'] == 30
    assert results['time_in_ranges']['time_in_l1_hypoglycemia'] == 20
    assert results['time_in_ranges']['time_in_l2_hypoglycemia'] == 10
    assert results['time_in_ranges']['time_in_hyperglycemia'] == 40
    assert results['time_in_ranges']['time_in_l1_hyperglycemia'] == 20
    assert results['time_in_ranges']['time_in_l2_hyperglycemia'] == 20

    # Risk fields
    assert type(results['risk']) is dict
    assert np.round(results['risk']['adrr']*100)/100 == 61.13
    assert np.round(results['risk']['lbgi']*100)/100 == 6.76
    assert np.round(results['risk']['hbgi']*100)/100 == 7.57
    assert np.round(results['risk']['bgri']*100)/100 == 14.32
    assert np.round(results['risk']['gri']*100)/100 == 100.00

    # Glycemic transformation fields
    assert type(results['glycemic_transformation']) is dict
    assert np.round(results['glycemic_transformation']['grade_score'] * 100) / 100 == 13.03
    assert np.round(results['glycemic_transformation']['grade_hypo_score'] * 100) / 100 == 38.99
    assert np.round(results['glycemic_transformation']['grade_hyper_score'] * 100) / 100 == 53.84
    assert np.round(results['glycemic_transformation']['grade_eu_score'] * 100) / 100 == 7.17
    assert np.round(results['glycemic_transformation']['igc'] * 100) / 100 == 4.67
    assert np.round(results['glycemic_transformation']['hypo_index'] * 100) / 100 == 3.67
    assert np.round(results['glycemic_transformation']['hyper_index'] * 100) / 100 == 1.01
    assert np.round(results['glycemic_transformation']['mr_index'] * 100) / 100 == 28.92

    # Pregnancy
    agata = Agata(glycemic_target='pregnancy')

    results = agata.analyze_glucose_profile(data=data)
    assert type(results) is dict

    # Variability metrics
    assert type(results['variability']) is dict
    assert np.round(results['variability']['mean_glucose']*1000)/1000 == 143.000
    assert np.round(results['variability']['median_glucose']*1000)/1000 == 135.000
    assert np.round(results['variability']['std_glucose']*1000)/1000 == 83.540
    assert np.round(results['variability']['cv_glucose']*1000)/1000 == 58.419
    assert np.round(results['variability']['range_glucose']*1000)/1000 == 220.000
    assert np.round(results['variability']['iqr_glucose']*1000)/1000 == 135.000
    assert np.round(results['variability']['auc_glucose']*1000)/1000 == 7150.000
    assert np.round(results['variability']['gmi']*1000)/1000 == 6.731
    assert np.round(results['variability']['cogi']*1000)/1000 == 19.091
    assert np.isnan(np.round(results['variability']['conga']*1000)/1000)
    assert np.round(results['variability']['j_index']*1000)/1000 == 51.320
    assert np.round(results['variability']['mage_plus_index']*1000)/1000 == 0.0
    assert np.round(results['variability']['mage_minus_index']*1000)/1000 == 0.0
    assert np.round(results['variability']['mage_index']*1000)/1000 == 0.0
    assert np.round(results['variability']['ef_index']*1000)/1000 == 0.0
    assert np.isnan(np.round(results['variability']['modd']*1000)/1000)
    assert np.isnan(np.round(results['variability']['sddm_index']*1000)/1000)
    assert np.round(results['variability']['sdw_index']*1000)/1000 == 83.540
    assert np.round(results['variability']['std_glucose_roc']*1000)/1000 == 1.925
    assert np.round(results['variability']['cvga']*1000)/1000 == 4759.436

    # Time fields
    assert type(results['time_in_ranges']) is dict
    assert results['time_in_ranges']['time_in_target'] == 20
    assert results['time_in_ranges']['time_in_tight_target'] == 20
    assert results['time_in_ranges']['time_in_hypoglycemia'] == 30
    assert results['time_in_ranges']['time_in_l1_hypoglycemia'] == 20
    assert results['time_in_ranges']['time_in_l2_hypoglycemia'] == 10
    assert results['time_in_ranges']['time_in_hyperglycemia'] == 50
    assert results['time_in_ranges']['time_in_l1_hyperglycemia'] == 30
    assert results['time_in_ranges']['time_in_l2_hyperglycemia'] == 20

    # Risk fields
    assert type(results['risk']) is dict
    assert np.round(results['risk']['adrr'] * 100) / 100 == 61.13
    assert np.round(results['risk']['lbgi'] * 100) / 100 == 6.76
    assert np.round(results['risk']['hbgi'] * 100) / 100 == 7.57
    assert np.round(results['risk']['bgri'] * 100) / 100 == 14.32
    assert np.round(results['risk']['gri'] * 100) / 100 == 100.00

    # Glycemic transformation fields
    assert type(results['glycemic_transformation']) is dict
    assert np.round(results['glycemic_transformation']['grade_score'] * 100) / 100 == 13.03
    assert np.round(results['glycemic_transformation']['grade_hypo_score'] * 100) / 100 == 38.99
    assert np.round(results['glycemic_transformation']['grade_hyper_score'] * 100) / 100 == 53.84
    assert np.round(results['glycemic_transformation']['grade_eu_score'] * 100) / 100 == 7.17
    assert np.round(results['glycemic_transformation']['igc'] * 100) / 100 == 4.67
    assert np.round(results['glycemic_transformation']['hypo_index'] * 100) / 100 == 3.67
    assert np.round(results['glycemic_transformation']['hyper_index'] * 100) / 100 == 1.01
    assert np.round(results['glycemic_transformation']['mr_index'] * 100) / 100 == 28.92

    # Data quality fields
    assert type(results['data_quality']) is dict
    assert np.round(results['data_quality']['number_days_of_observation'] * 100) / 100 == 0.03
    assert np.round(results['data_quality']['missing_glucose_percentage'] * 100) / 100 == 9.09

    # Events fields
    # assert type(results['events']) is dict
    assert type(results['events']['hypoglycemic_events']) is dict
    assert type(results['events']['hyperglycemic_events']) is dict
    assert type(results['events']['extended_hypoglycemic_events']) is dict



import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.input_validator import check_str_parameter

import pytest

def test_check_str_parameter():
    """
    Unit test of check_str_parameter function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Test not error
    assert check_str_parameter("ciao")

    # Test errors
    with pytest.raises(Exception):
        check_str_parameter(50)


import numpy as np
from statsmodels.stats.diagnostic import lilliefors
from scipy.stats import ttest_ind, wilcoxon, mannwhitneyu,ranksums

from py_agata.variability import *
from py_agata.time_in_ranges import *
from py_agata.risk import *
from py_agata.glycemic_transformation import *
from py_agata.inspection import *

class Agata:
    """
    Core class of AGATA.

    ...
    Attributes
    ----------
    glycemic_target: str
        A string defining the set of glycemic targets to use.

    Methods
    -------
    analyze_glucose_profile():
        Runs ReplayBG.
    """

    def __init__(self, glycemic_target='diabetes'):
        self.glycemic_target = glycemic_target

    def analyze_glucose_profile(self, data):
        """
        Analyzes a single glucose profile.

        Parameters
        ----------
        data: pd.DataFrame
            Pandas dataframe with a column `glucose` containing the glucose data to analyze (in mg/dl).

        Returns
        -------
        results: dict
            A dictionary containing the results of the analysis i.e.:
            - variability: dict
                A dictionary containing the values of the variability related metrics.
            - time_in_ranges: dict
                A dictionary containing the values of the time in range related metrics.
            - risk: dict
                A dictionary containing the values of the risk related metrics.
            - glycemic_transformation: dict
                A dictionary containing the values of the glycemic_transformation related metrics.
            - data_quality: dict
                A dictionary containing the values of the data quality related metrics.
            - events: dict
                A dictionary containing the values of the events related metrics.


        Raises
        ------
        None

        See Also
        --------
        None

        Examples
        --------
        None

        References
        ----------
        None
        """
        # Check input
        check_dataframe(data)
        check_data_columns(data)
        check_homogeneous_timegrid(data)

        results = dict()

        # Get variability metrics
        results['variability'] = dict()
        results['variability']['mean_glucose'] = mean_glucose(data)
        results['variability']['median_glucose'] = median_glucose(data)
        results['variability']['std_glucose'] = std_glucose(data)
        results['variability']['cv_glucose'] = cv_glucose(data)
        results['variability']['range_glucose'] = range_glucose(data)
        results['variability']['iqr_glucose'] = iqr_glucose(data)
        results['variability']['auc_glucose'] = auc_glucose(data)
        results['variability']['gmi'] = gmi(data)
        results['variability']['cogi'] = cogi(data)
        results['variability']['conga'] = conga(data)
        results['variability']['j_index'] = j_index(data)
        results['variability']['mage_plus_index'] = mage_plus_index(data)
        results['variability']['mage_minus_index'] = mage_minus_index(data)
        results['variability']['mage_index'] = mage_index(data)
        results['variability']['ef_index'] = ef_index(data)
        results['variability']['modd'] = modd(data)
        results['variability']['sddm_index'] = sddm_index(data)
        results['variability']['sdw_index'] = sdw_index(data)
        results['variability']['std_glucose_roc'] = std_glucose_roc(data)
        results['variability']['cvga'] = cvga(data)

        # Get time metrics
        results['time_in_ranges'] = dict()
        results['time_in_ranges']['time_in_target'] = time_in_target(data, self.glycemic_target)
        results['time_in_ranges']['time_in_tight_target'] = time_in_tight_target(data, self.glycemic_target)
        results['time_in_ranges']['time_in_hypoglycemia'] = time_in_hypoglycemia(data, self.glycemic_target)
        results['time_in_ranges']['time_in_l1_hypoglycemia'] = time_in_l1_hypoglycemia(data, self.glycemic_target)
        results['time_in_ranges']['time_in_l2_hypoglycemia'] = time_in_l2_hypoglycemia(data, self.glycemic_target)
        results['time_in_ranges']['time_in_hyperglycemia'] = time_in_hyperglycemia(data, self.glycemic_target)
        results['time_in_ranges']['time_in_l1_hyperglycemia'] = time_in_l1_hyperglycemia(data, self.glycemic_target)
        results['time_in_ranges']['time_in_l2_hyperglycemia'] = time_in_l2_hyperglycemia(data, self.glycemic_target)

        # Get risk metrics
        results['risk'] = dict()
        results['risk']['adrr'] = adrr(data)
        results['risk']['lbgi'] = lbgi(data)
        results['risk']['hbgi'] = hbgi(data)
        results['risk']['bgri'] = bgri(data)
        results['risk']['gri'] = gri(data)

        # Get glycemic transformation metrics
        results['glycemic_transformation'] = dict()
        results['glycemic_transformation']['grade_score'] = grade_score(data)
        results['glycemic_transformation']['grade_hypo_score'] = grade_hypo_score(data)
        results['glycemic_transformation']['grade_hyper_score'] = grade_hyper_score(data)
        results['glycemic_transformation']['grade_eu_score'] = grade_eu_score(data)
        results['glycemic_transformation']['igc'] = igc(data)
        results['glycemic_transformation']['hypo_index'] = hypo_index(data)
        results['glycemic_transformation']['hyper_index'] = hyper_index(data)
        results['glycemic_transformation']['mr_index'] = mr_index(data)

        # Event metrics
        results['events'] = dict()
        results['events']['hypoglycemic_events'] = find_hypoglycemic_events_by_level(data, glycemic_target=self.glycemic_target)
        results['events']['hyperglycemic_events'] = find_hyperglycemic_events_by_level(data, glycemic_target=self.glycemic_target)
        results['events']['extended_hypoglycemic_events'] = find_extended_hypoglycemic_events(data)

        # Data quality metrics
        results['data_quality'] = dict()
        results['data_quality']['number_days_of_observation'] = number_days_of_observation(data)
        results['data_quality']['missing_glucose_percentage'] = missing_glucose_percentage(data)

        # Return results
        return results

    def analyze_one_arm(self, data):
        """
        Analyzes glucose data of one arm.

        Parameters
        ----------
        data: list of pd.DataFrame
            List of pandas dataframes with a column `glucose` containing the glucose data
            to analyze (in mg/dl).

        Returns
        -------
        results: dict
            A dictionary containing the results of the analysis i.e.:
            - variability: dict
                A dictionary containing the values of the variability related metrics.
            - time_in_ranges: dict
                A dictionary containing the values of the time in range related metrics.
            - risk: dict
                A dictionary containing the values of the risk related metrics.
            - glycemic_transformation: dict
                A dictionary containing the values of the glycemic_transformation related metrics.
            - data_quality: dict
                A dictionary containing the values of the data quality related metrics.
            - events: dict
                A dictionary containing the values of the events related metrics.


        Raises
        ------
        None

        See Also
        --------
        None

        Examples
        --------
        None

        References
        ----------
        None
        """
        for d in data:
            # Check input
            check_dataframe(d)
            check_data_columns(d)
            check_homogeneous_timegrid(d)

        results = dict()


        # Variability
        results["variability"] = dict()
        metric_list = [mean_glucose, median_glucose, std_glucose, cv_glucose, range_glucose, iqr_glucose,
                       auc_glucose, gmi, cogi, conga, j_index, mage_index, mage_minus_index, mage_plus_index,
                       ef_index, modd, sddm_index, sdw_index, std_glucose_roc, cvga]
        metric_list_name = ['mean_glucose', 'median_glucose', 'std_glucose', 'cv_glucose', 'range_glucose',
                            'iqr_glucose',
                            'auc_glucose', 'gmi', 'cogi', 'conga', 'j_index', 'mage_index', 'mage_minus_index',
                            'mage_plus_index',
                            'ef_index', 'modd', 'sddm_index', 'sdw_index', 'std_glucose_roc', 'cvga']

        for m in range(len(metric_list_name)):
            results["variability"][metric_list_name[m]] = dict()
            results["variability"][metric_list_name[m]]["values"] = np.zeros(shape=(len(data),))
            for d in range(len(data)):
                results["variability"][metric_list_name[m]]["values"][d] = metric_list[m](data[d])
            results["variability"][metric_list_name[m]]["mean"] = np.nanmean(results["variability"][metric_list_name[m]]["values"])
            results["variability"][metric_list_name[m]]["std"] = np.nanstd(results["variability"][metric_list_name[m]]["values"])
            results["variability"][metric_list_name[m]]["median"] = np.nanmedian(results["variability"][metric_list_name[m]]["values"])
            results["variability"][metric_list_name[m]]["prc_5"] = np.nanpercentile(results["variability"][metric_list_name[m]]["values"], 5)
            results["variability"][metric_list_name[m]]["prc_25"] = np.nanpercentile(results["variability"][metric_list_name[m]]["values"], 25)
            results["variability"][metric_list_name[m]]["prc_75"] = np.nanpercentile(results["variability"][metric_list_name[m]]["values"], 75)
            results["variability"][metric_list_name[m]]["prc_95"] = np.nanpercentile(results["variability"][metric_list_name[m]]["values"], 95)


        # Time in ranges
        results["time_in_ranges"] = dict()
        metric_list = [time_in_target, time_in_tight_target, time_in_hypoglycemia,
                       time_in_l1_hypoglycemia, time_in_l2_hypoglycemia, time_in_hyperglycemia,
                       time_in_l1_hyperglycemia, time_in_l2_hyperglycemia]
        metric_list_name = ['time_in_target', 'time_in_tight_target', 'time_in_hypoglycemia',
                            'time_in_l1_hypoglycemia', 'time_in_l2_hypoglycemia', 'time_in_hyperglycemia',
                            'time_in_l1_hyperglycemia', 'time_in_l2_hyperglycemia']

        for m in range(len(metric_list_name)):
            results["time_in_ranges"][metric_list_name[m]] = dict()
            results["time_in_ranges"][metric_list_name[m]]["values"] = np.zeros(shape=(len(data),))
            for d in range(len(data)):
                results["time_in_ranges"][metric_list_name[m]]["values"][d] = metric_list[m](data[d], self.glycemic_target)
            results["time_in_ranges"][metric_list_name[m]]["mean"] = np.nanmean(
                results["time_in_ranges"][metric_list_name[m]]["values"])
            results["time_in_ranges"][metric_list_name[m]]["std"] = np.nanstd(
                results["time_in_ranges"][metric_list_name[m]]["values"])
            results["time_in_ranges"][metric_list_name[m]]["median"] = np.nanmedian(
                results["time_in_ranges"][metric_list_name[m]]["values"])
            results["time_in_ranges"][metric_list_name[m]]["prc_5"] = np.nanpercentile(
                results["time_in_ranges"][metric_list_name[m]]["values"], 5)
            results["time_in_ranges"][metric_list_name[m]]["prc_25"] = np.nanpercentile(
                results["time_in_ranges"][metric_list_name[m]]["values"], 25)
            results["time_in_ranges"][metric_list_name[m]]["prc_75"] = np.nanpercentile(
                results["time_in_ranges"][metric_list_name[m]]["values"], 75)
            results["time_in_ranges"][metric_list_name[m]]["prc_95"] = np.nanpercentile(
                results["time_in_ranges"][metric_list_name[m]]["values"], 95)

        # Risk
        results["risk"] = dict()
        metric_list = [adrr, lbgi, hbgi, bgri, gri]
        metric_list_name = ['adrr', 'lbgi', 'hbgi', 'bgri', 'gri']

        for m in range(len(metric_list_name)):
            results["risk"][metric_list_name[m]] = dict()
            results["risk"][metric_list_name[m]]["values"] = np.zeros(shape=(len(data),))
            for d in range(len(data)):
                results["risk"][metric_list_name[m]]["values"][d] = metric_list[m](data[d])
            results["risk"][metric_list_name[m]]["mean"] = np.nanmean(
                results["risk"][metric_list_name[m]]["values"])
            results["risk"][metric_list_name[m]]["std"] = np.nanstd(
                results["risk"][metric_list_name[m]]["values"])
            results["risk"][metric_list_name[m]]["median"] = np.nanmedian(
                results["risk"][metric_list_name[m]]["values"])
            results["risk"][metric_list_name[m]]["prc_5"] = np.nanpercentile(
                results["risk"][metric_list_name[m]]["values"], 5)
            results["risk"][metric_list_name[m]]["prc_25"] = np.nanpercentile(
                results["risk"][metric_list_name[m]]["values"], 25)
            results["risk"][metric_list_name[m]]["prc_75"] = np.nanpercentile(
                results["risk"][metric_list_name[m]]["values"], 75)
            results["risk"][metric_list_name[m]]["prc_95"] = np.nanpercentile(
                results["risk"][metric_list_name[m]]["values"], 95)

        # Glycemic transformation
        results["glycemic_transformation"] = dict()
        metric_list = [grade_score, grade_hypo_score, grade_hyper_score,
                       grade_eu_score, igc, hypo_index,
                       hyper_index, mr_index]
        metric_list_name = ['grade_score', 'grade_hypo_score', 'grade_hyper_score',
                            'grade_eu_score', 'igc', 'hypo_index','hyper_index',
                            'mr_index']

        for m in range(len(metric_list_name)):
            results["glycemic_transformation"][metric_list_name[m]] = dict()
            results["glycemic_transformation"][metric_list_name[m]]["values"] = np.zeros(shape=(len(data),))
            for d in range(len(data)):
                results["glycemic_transformation"][metric_list_name[m]]["values"][d] = metric_list[m](data[d])
            results["glycemic_transformation"][metric_list_name[m]]["mean"] = np.nanmean(
                results["glycemic_transformation"][metric_list_name[m]]["values"])
            results["glycemic_transformation"][metric_list_name[m]]["std"] = np.nanstd(
                results["glycemic_transformation"][metric_list_name[m]]["values"])
            results["glycemic_transformation"][metric_list_name[m]]["median"] = np.nanmedian(
                results["glycemic_transformation"][metric_list_name[m]]["values"])
            results["glycemic_transformation"][metric_list_name[m]]["prc_5"] = np.nanpercentile(
                results["glycemic_transformation"][metric_list_name[m]]["values"], 5)
            results["glycemic_transformation"][metric_list_name[m]]["prc_25"] = np.nanpercentile(
                results["glycemic_transformation"][metric_list_name[m]]["values"], 25)
            results["glycemic_transformation"][metric_list_name[m]]["prc_75"] = np.nanpercentile(
                results["glycemic_transformation"][metric_list_name[m]]["values"], 75)
            results["glycemic_transformation"][metric_list_name[m]]["prc_95"] = np.nanpercentile(
                results["glycemic_transformation"][metric_list_name[m]]["values"], 95)

        # Data quality
        results["data_quality"] = dict()
        metric_list = [number_days_of_observation, missing_glucose_percentage]
        metric_list_name = ['number_days_of_observation','missing_glucose_percentage']

        for m in range(len(metric_list_name)):
            results["data_quality"][metric_list_name[m]] = dict()
            results["data_quality"][metric_list_name[m]]["values"] = np.zeros(shape=(len(data),))
            for d in range(len(data)):
                results["data_quality"][metric_list_name[m]]["values"][d] = metric_list[m](data[d])
            results["data_quality"][metric_list_name[m]]["mean"] = np.nanmean(
                results["data_quality"][metric_list_name[m]]["values"])
            results["data_quality"][metric_list_name[m]]["std"] = np.nanstd(
                results["data_quality"][metric_list_name[m]]["values"])
            results["data_quality"][metric_list_name[m]]["median"] = np.nanmedian(
                results["data_quality"][metric_list_name[m]]["values"])
            results["data_quality"][metric_list_name[m]]["prc_5"] = np.nanpercentile(
                results["data_quality"][metric_list_name[m]]["values"], 5)
            results["data_quality"][metric_list_name[m]]["prc_25"] = np.nanpercentile(
                results["data_quality"][metric_list_name[m]]["values"], 25)
            results["data_quality"][metric_list_name[m]]["prc_75"] = np.nanpercentile(
                results["data_quality"][metric_list_name[m]]["values"], 75)
            results["data_quality"][metric_list_name[m]]["prc_95"] = np.nanpercentile(
                results["data_quality"][metric_list_name[m]]["values"], 95)

        # Events
        results["events"] = dict()
        results["events"]["hyperglycemic_events"] = dict()
        results["events"]["hyperglycemic_events"]["hyper"] = dict()
        results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"] = dict()
        results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"] = dict()
        results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hyperglycemic_events"]["l1"] = dict()
        results["events"]["hyperglycemic_events"]["l1"]["mean_duration"] = dict()
        results["events"]["hyperglycemic_events"]["l1"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hyperglycemic_events"]["l1"]["events_per_week"] = dict()
        results["events"]["hyperglycemic_events"]["l1"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hyperglycemic_events"]["l2"] = dict()
        results["events"]["hyperglycemic_events"]["l2"]["mean_duration"] = dict()
        results["events"]["hyperglycemic_events"]["l2"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hyperglycemic_events"]["l2"]["events_per_week"] = dict()
        results["events"]["hyperglycemic_events"]["l2"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))

        results["events"]["hypoglycemic_events"] = dict()
        results["events"]["hypoglycemic_events"]["hypo"] = dict()
        results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"] = dict()
        results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"] = dict()
        results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hypoglycemic_events"]["l1"] = dict()
        results["events"]["hypoglycemic_events"]["l1"]["mean_duration"] = dict()
        results["events"]["hypoglycemic_events"]["l1"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hypoglycemic_events"]["l1"]["events_per_week"] = dict()
        results["events"]["hypoglycemic_events"]["l1"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hypoglycemic_events"]["l2"] = dict()
        results["events"]["hypoglycemic_events"]["l2"]["mean_duration"] = dict()
        results["events"]["hypoglycemic_events"]["l2"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["hypoglycemic_events"]["l2"]["events_per_week"] = dict()
        results["events"]["hypoglycemic_events"]["l2"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))

        results["events"]["extended_hypoglycemic_events"] = dict()
        results["events"]["extended_hypoglycemic_events"]["mean_duration"] = dict()
        results["events"]["extended_hypoglycemic_events"]["mean_duration"]["values"] = np.zeros(shape=(len(data),))
        results["events"]["extended_hypoglycemic_events"]["events_per_week"] = dict()
        results["events"]["extended_hypoglycemic_events"]["events_per_week"]["values"] = np.zeros(shape=(len(data),))

        for d in range(len(data)):
            r = find_hyperglycemic_events_by_level(data[d], glycemic_target=self.glycemic_target)
            results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"]["values"][d] = r["hyper"]["mean_duration"]
            results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"]["values"][d] = r["hyper"]["events_per_week"]
            results["events"]["hyperglycemic_events"]["l1"]["mean_duration"]["values"][d] = r["l1"]["mean_duration"]
            results["events"]["hyperglycemic_events"]["l1"]["events_per_week"]["values"][d] = r["l1"]["events_per_week"]
            results["events"]["hyperglycemic_events"]["l2"]["mean_duration"]["values"][d] = r["l2"]["mean_duration"]
            results["events"]["hyperglycemic_events"]["l2"]["events_per_week"]["values"][d] = r["l2"]["events_per_week"]

            r = find_hypoglycemic_events_by_level(data[d], glycemic_target=self.glycemic_target)
            results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"]["values"][d] = r["hypo"]["mean_duration"]
            results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"]["values"][d] = r["hypo"]["events_per_week"]
            results["events"]["hypoglycemic_events"]["l1"]["mean_duration"]["values"][d] = r["l1"]["mean_duration"]
            results["events"]["hypoglycemic_events"]["l1"]["events_per_week"]["values"][d] = r["l1"]["events_per_week"]
            results["events"]["hypoglycemic_events"]["l2"]["mean_duration"]["values"][d] = r["l2"]["mean_duration"]
            results["events"]["hypoglycemic_events"]["l2"]["events_per_week"]["values"][d] = r["l2"]["events_per_week"]

            r = find_extended_hypoglycemic_events(data[d])
            results["events"]["extended_hypoglycemic_events"]["mean_duration"]["values"][d] = r["mean_duration"]
            results["events"]["extended_hypoglycemic_events"]["events_per_week"]["values"][d] = r["events_per_week"]

        stat = ["mean", "std", "median"]
        f_stat = [np.nanmean, np.nanstd, np.nanmedian]
        prc = ["prc_5","prc_25","prc_75","prc_95"]
        qs = [5, 25, 75, 95]
        for s in range(len(stat)):
            results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"]["values"])
            results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"]["values"])
            results["events"]["hyperglycemic_events"]["l1"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["l1"]["mean_duration"]["values"])
            results["events"]["hyperglycemic_events"]["l1"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["l1"]["events_per_week"]["values"])
            results["events"]["hyperglycemic_events"]["l2"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["l2"]["mean_duration"]["values"])
            results["events"]["hyperglycemic_events"]["l2"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hyperglycemic_events"]["l2"]["events_per_week"]["values"])

            results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"]["values"])
            results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"]["values"])
            results["events"]["hypoglycemic_events"]["l1"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["l1"]["mean_duration"]["values"])
            results["events"]["hypoglycemic_events"]["l1"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["l1"]["events_per_week"]["values"])
            results["events"]["hypoglycemic_events"]["l2"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["l2"]["mean_duration"]["values"])
            results["events"]["hypoglycemic_events"]["l2"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["hypoglycemic_events"]["l2"]["events_per_week"]["values"])

            results["events"]["extended_hypoglycemic_events"]["mean_duration"][stat[s]] = f_stat[s](results["events"]["extended_hypoglycemic_events"]["mean_duration"]["values"])
            results["events"]["extended_hypoglycemic_events"]["events_per_week"][stat[s]] = f_stat[s](results["events"]["extended_hypoglycemic_events"]["events_per_week"]["values"])

        for s in range(len(prc)):
            results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["hyper"]["mean_duration"]["values"], qs[s])
            results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["hyper"]["events_per_week"]["values"], qs[s])
            results["events"]["hyperglycemic_events"]["l1"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["l1"]["mean_duration"]["values"], qs[s])
            results["events"]["hyperglycemic_events"]["l1"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["l1"]["events_per_week"]["values"], qs[s])
            results["events"]["hyperglycemic_events"]["l2"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["l2"]["mean_duration"]["values"], qs[s])
            results["events"]["hyperglycemic_events"]["l2"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hyperglycemic_events"]["l2"]["events_per_week"]["values"], qs[s])

            results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["hypo"]["mean_duration"]["values"], qs[s])
            results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["hypo"]["events_per_week"]["values"], qs[s])
            results["events"]["hypoglycemic_events"]["l1"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["l1"]["mean_duration"]["values"], qs[s])
            results["events"]["hypoglycemic_events"]["l1"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["l1"]["events_per_week"]["values"], qs[s])
            results["events"]["hypoglycemic_events"]["l2"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["l2"]["mean_duration"]["values"], qs[s])
            results["events"]["hypoglycemic_events"]["l2"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["hypoglycemic_events"]["l2"]["events_per_week"]["values"], qs[s])

            results["events"]["extended_hypoglycemic_events"]["mean_duration"][prc[s]] = np.nanpercentile(results["events"]["extended_hypoglycemic_events"]["mean_duration"]["values"], qs[s])
            results["events"]["extended_hypoglycemic_events"]["events_per_week"][prc[s]] = np.nanpercentile(results["events"]["extended_hypoglycemic_events"]["events_per_week"]["values"], qs[s])

        return results

    def compare_two_arms(self, arm_1, arm_2, is_paired, alpha):
        """
        Analyzes and compares glucose data of one arm.

        Parameters
        ----------
        arm_1: list of pd.DataFrame
            List of pandas dataframes with a column `glucose` containing the glucose data
            to analyze (in mg/dl). These are the data of the first arm
        arm_2: list of pd.DataFrame
            List of pandas dataframes with a column `glucose` containing the glucose data
            to analyze (in mg/dl). These are the data of the second arm
         is_paired: bool
            A boolean flag defining whether to run paired or unpaired analysis. Commonly paired tests are performed
            when data of the same patients are present in both arms, unpaired otherwise
        alpha: float
            The significance level to use during the statistical analysis

        Returns
        -------
        results: dict
            A dictionary containing two dictionaries (`arm_1` and `arm_2`) each containing the results of the analysis
             of the two arms with fields:
            - variability: dict
                A dictionary containing the values of the variability related metrics.
            - time_in_ranges: dict
                A dictionary containing the values of the time in range related metrics.
            - risk: dict
                A dictionary containing the values of the risk related metrics.
            - glycemic_transformation: dict
                A dictionary containing the values of the glycemic_transformation related metrics.
            - data_quality: dict
                A dictionary containing the values of the data quality related metrics.
            - events: dict
                A dictionary containing the values of the events related metrics.

         stats: dict
            A structure that contains for each of the considered metrics the result of the statistical test with
            field `p` (p-value value) and `h` null hypothesis accepted or rejected. Statistical tests are:
                - t-test if the test `is_paired` and the samples are both gaussian distributed
                (checked with the Lilliefors test)
                - unpaired t-test if the test not `is_paired` and the samples are both gaussian distributed
                (checked with the Lilliefors test)
                - Wilcoxon rank test if the test `is_paired` and at least one of the samples is not gaussian distributed
                (checked with the Lilliefors test)
                - Mann-Whitney U-test if the test not `is_paired` and at least one of the samples is not gaussian
                distributed (checked with the Lilliefors test).
        Raises
        ------
        None

        See Also
        --------
        None

        Examples
        --------
        None

        References
        ----------
        None
        """
        for d in arm_1:
            # Check input
            check_dataframe(d)
            check_data_columns(d)
            check_homogeneous_timegrid(d)
        for d in arm_2:
            # Check input
            check_dataframe(d)
            check_data_columns(d)
            check_homogeneous_timegrid(d)

        results = dict()
        results["arm_1"] = self.analyze_one_arm(arm_1)
        results["arm_2"] = self.analyze_one_arm(arm_2)
        stats = dict()

        # Variability
        stats["variability"] = dict()
        metric_list_name = ['mean_glucose', 'median_glucose', 'std_glucose', 'cv_glucose', 'range_glucose',
                            'iqr_glucose',
                            'auc_glucose', 'gmi', 'cogi', 'conga', 'j_index', 'mage_index', 'mage_minus_index',
                            'mage_plus_index',
                            'ef_index', 'modd', 'sddm_index', 'sdw_index', 'std_glucose_roc', 'cvga']

        for m in metric_list_name:
            r1 = results["arm_1"]["variability"][m]["values"]
            r2 = results["arm_2"]["variability"][m]["values"]

            # Correct for lilliefors behaviour
            if(r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0

            stats["variability"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or ((p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["variability"][m]["p"] = t.pvalue
                if np.isnan(stats["variability"][m]["p"]):
                    stats["variability"][m]["h"] = np.nan
                else:
                    stats["variability"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["variability"][m]["h"] = 0
                        stats["variability"][m]["p"] = 1
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["variability"][m]["p"] = t.pvalue
                        if np.isnan(stats["variability"][m]["p"]):
                            stats["variability"][m]["h"] = np.nan
                        else:
                            stats["variability"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["variability"][m]["p"] = t.pvalue
                    if np.isnan(stats["variability"][m]["p"]):
                        stats["variability"][m]["h"] = np.nan
                    else:
                        stats["variability"][m]["h"] = 1 * (t.pvalue < alpha)

        # Time in ranges
        stats["time_in_ranges"] = dict()
        metric_list_name = ['time_in_target', 'time_in_tight_target', 'time_in_hypoglycemia',
                            'time_in_l1_hypoglycemia', 'time_in_l2_hypoglycemia', 'time_in_hyperglycemia',
                            'time_in_l1_hyperglycemia', 'time_in_l2_hyperglycemia']

        for m in metric_list_name:
            r1 = results["arm_1"]["time_in_ranges"][m]["values"]
            r2 = results["arm_2"]["time_in_ranges"][m]["values"]

            # Correct for lilliefors behaviour
            if (r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0

            stats["time_in_ranges"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or (
                    (p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["time_in_ranges"][m]["p"] = t.pvalue
                if np.isnan(stats["time_in_ranges"][m]["p"]):
                    stats["time_in_ranges"][m]["h"] = np.nan
                else:
                    stats["time_in_ranges"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["time_in_ranges"][m]["h"] = 0
                        stats["time_in_ranges"][m]["p"] = 1
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["time_in_ranges"][m]["p"] = t.pvalue
                        if np.isnan(stats["time_in_ranges"][m]["p"]):
                            stats["time_in_ranges"][m]["h"] = np.nan
                        else:
                            stats["time_in_ranges"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["time_in_ranges"][m]["p"] = t.pvalue
                    if np.isnan(stats["time_in_ranges"][m]["p"]):
                        stats["time_in_ranges"][m]["h"] = np.nan
                    else:
                        stats["time_in_ranges"][m]["h"] = 1 * (t.pvalue < alpha)
        # Risk
        stats["risk"] = dict()
        metric_list_name = ['adrr', 'lbgi', 'hbgi', 'bgri', 'gri']

        for m in metric_list_name:
            r1 = results["arm_1"]["risk"][m]["values"]
            r2 = results["arm_2"]["risk"][m]["values"]

            # Correct for lilliefors behaviour
            if (r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0

            stats["risk"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or (
                    (p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["risk"][m]["p"] = t.pvalue
                if np.isnan(stats["risk"][m]["p"]):
                    stats["risk"][m]["h"] = np.nan
                else:
                    stats["risk"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["risk"][m]["h"] = 0
                        stats["risk"][m]["p"] = 1
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["risk"][m]["p"] = t.pvalue
                        if np.isnan(stats["risk"][m]["p"]):
                            stats["risk"][m]["h"] = np.nan
                        else:
                            stats["risk"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["risk"][m]["p"] = t.pvalue
                    if np.isnan(stats["risk"][m]["p"]):
                        stats["risk"][m]["h"] = np.nan
                    else:
                        stats["risk"][m]["h"] = 1 * (t.pvalue < alpha)

        # Glycemic transformation
        stats["glycemic_transformation"] = dict()
        metric_list_name = ['grade_score', 'grade_hypo_score', 'grade_hyper_score',
                                        'grade_eu_score', 'igc', 'hypo_index', 'hyper_index',
                                        'mr_index']

        for m in metric_list_name:
            r1 = results["arm_1"]["glycemic_transformation"][m]["values"]
            r2 = results["arm_2"]["glycemic_transformation"][m]["values"]

            # Correct for lilliefors behaviour
            if (r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0

            stats["glycemic_transformation"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or (
                    (p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["glycemic_transformation"][m]["p"] = t.pvalue
                if np.isnan(stats["glycemic_transformation"][m]["p"]):
                    stats["glycemic_transformation"][m]["h"] = np.nan
                else:
                    stats["glycemic_transformation"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["glycemic_transformation"][m]["h"] = 0
                        stats["glycemic_transformation"][m]["p"] = 1
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["glycemic_transformation"][m]["p"] = t.pvalue
                        if np.isnan(stats["glycemic_transformation"][m]["p"]):
                            stats["glycemic_transformation"][m]["h"] = np.nan
                        else:
                            stats["glycemic_transformation"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["glycemic_transformation"][m]["p"] = t.pvalue
                    if np.isnan(stats["glycemic_transformation"][m]["p"]):
                        stats["glycemic_transformation"][m]["h"] = np.nan
                    else:
                        stats["glycemic_transformation"][m]["h"] = 1 * (t.pvalue < alpha)

        # Data quality
        stats["data_quality"] = dict()
        metric_list_name = ['number_days_of_observation','missing_glucose_percentage']

        for m in metric_list_name:
            r1 = results["arm_1"]["data_quality"][m]["values"]
            r2 = results["arm_2"]["data_quality"][m]["values"]

            # Correct for lilliefors behaviour
            if (r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0

            stats["data_quality"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or (
                    (p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["data_quality"][m]["p"] = t.pvalue
                if np.isnan(stats["data_quality"][m]["p"]):
                    stats["data_quality"][m]["h"] = np.nan
                else:
                    stats["data_quality"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["data_quality"][m]["h"] = 0
                        stats["data_quality"][m]["p"] = 1
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["data_quality"][m]["p"] = t.pvalue
                        if np.isnan(stats["data_quality"][m]["p"]):
                            stats["data_quality"][m]["h"] = np.nan
                        else:
                            stats["data_quality"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["data_quality"][m]["p"] = t.pvalue
                    if np.isnan(stats["data_quality"][m]["p"]):
                        stats["data_quality"][m]["h"] = np.nan
                    else:
                        stats["data_quality"][m]["h"] = 1 * (t.pvalue < alpha)

        # Events
        stats["events"] = dict()
        stats["events"]["hypoglycemic_events"] = dict()
        stats["events"]["hypoglycemic_events"]["hypo"] = dict()
        stats["events"]["hypoglycemic_events"]["l1"] = dict()
        stats["events"]["hypoglycemic_events"]["l2"] = dict()
        stats["events"]["hyperglycemic_events"] = dict()
        stats["events"]["hyperglycemic_events"]["hyper"] = dict()
        stats["events"]["hyperglycemic_events"]["l1"] = dict()
        stats["events"]["hyperglycemic_events"]["l2"] = dict()

        cat = ["hypoglycemic_events"]
        sub_cat = ['hypo', 'l1', 'l2']
        metric_list_name = ['mean_duration', 'events_per_week']

        for c in cat:
            for s in sub_cat:
                for m in metric_list_name:
                    r1 = results["arm_1"]["events"][c][s][m]["values"]
                    r2 = results["arm_2"]["events"][c][s][m]["values"]
                    # Correct for lilliefors behaviour
                    if (r1[~np.isnan(r1)].size >= 4):
                        ks, p1 = lilliefors(r1[~np.isnan(r1)])
                    else:
                        p1 = 0
                    if (r2[~np.isnan(r2)].size >= 4):
                        ks, p2 = lilliefors(r2[~np.isnan(r2)])
                    else:
                        p2 = 0

                    stats["events"][c][s][m] = dict()
                    if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or ((p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                        t = ttest_ind(r1, r2, nan_policy="omit")
                        stats["events"][c][s][m]["p"] = t.pvalue
                        if np.isnan(stats["events"][c][s][m]["p"]):
                            stats["events"][c][s][m]["h"] = np.nan
                        else:
                            stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)
                    else:
                        if is_paired:
                            idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                            if np.all(r1[idxs] - r2[idxs]) == 0:
                                stats["events"][c][s][m]["h"] = np.nan
                                stats["events"][c][s][m]["p"] = np.nan
                            else:
                                t = wilcoxon(r1, r2, nan_policy="omit")
                                stats["events"][c][s][m]["p"] = t.pvalue
                                if np.isnan(stats["events"][c][s][m]["p"]):
                                    stats["events"][c][s][m]["h"] = np.nan
                                else:
                                    stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)
                        else:
                            t = mannwhitneyu(r1, r2)
                            stats["events"][c][s][m]["p"] = t.pvalue
                            if np.isnan(stats["events"][c][s][m]["p"]):
                                stats["events"][c][s][m]["h"] = np.nan
                            else:
                                stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)

        cat = ["hyperglycemic_events"]
        sub_cat = ['hyper', 'l1', 'l2']
        metric_list_name = ['mean_duration', 'events_per_week']

        for c in cat:
            for s in sub_cat:
                for m in metric_list_name:
                    r1 = results["arm_1"]["events"][c][s][m]["values"]
                    r2 = results["arm_2"]["events"][c][s][m]["values"]
                    # Correct for lilliefors behaviour
                    if (r1[~np.isnan(r1)].size >= 4):
                        ks, p1 = lilliefors(r1[~np.isnan(r1)])
                    else:
                        p1 = 0
                    if (r2[~np.isnan(r2)].size >= 4):
                        ks, p2 = lilliefors(r2[~np.isnan(r2)])
                    else:
                        p2 = 0

                    stats["events"][c][s][m] = dict()
                    if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or ((p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                        t = ttest_ind(r1, r2, nan_policy="omit")
                        stats["events"][c][s][m]["p"] = t.pvalue
                        if np.isnan(stats["events"][c][s][m]["p"]):
                            stats["events"][c][s][m]["h"] = np.nan
                        else:
                            stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)
                    else:
                        if is_paired:
                            idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                            if np.all(r1[idxs] - r2[idxs]) == 0:
                                stats["events"][c][s][m]["h"] = np.nan
                                stats["events"][c][s][m]["p"] = np.nan
                            else:
                                t = wilcoxon(r1, r2, nan_policy="omit")
                                stats["events"][c][s][m]["p"] = t.pvalue
                                if np.isnan(stats["events"][c][s][m]["p"]):
                                    stats["events"][c][s][m]["h"] = np.nan
                                else:
                                    stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)
                        else:
                            t = mannwhitneyu(r1, r2)
                            stats["events"][c][s][m]["p"] = t.pvalue
                            if np.isnan(stats["events"][c][s][m]["p"]):
                                stats["events"][c][s][m]["h"] = np.nan
                            else:
                                stats["events"][c][s][m]["h"] = 1 * (t.pvalue < alpha)

        stats["events"]["extended_hypoglycemic_events"] = dict()
        for m in metric_list_name:
            r1 = results["arm_1"]["events"]["extended_hypoglycemic_events"][m]["values"]
            r2 = results["arm_2"]["events"]["extended_hypoglycemic_events"][m]["values"]
            # Correct for lilliefors behaviour
            if (r1[~np.isnan(r1)].size >= 4):
                ks, p1 = lilliefors(r1[~np.isnan(r1)])
            else:
                p1 = 0
            if (r2[~np.isnan(r2)].size >= 4):
                ks, p2 = lilliefors(r2[~np.isnan(r2)])
            else:
                p2 = 0
            stats["events"]["extended_hypoglycemic_events"][m] = dict()
            if np.sum(~np.isnan(r1)) < 4 or np.sum(~np.isnan(r2)) < 4 or ((p1 > 0.05 or np.isnan(p1)) and (p2 > 0.05 or np.isnan(p2))):
                t = ttest_ind(r1, r2, nan_policy="omit")
                stats["events"]["extended_hypoglycemic_events"][m]["p"] = t.pvalue
                if np.isnan(stats["events"]["extended_hypoglycemic_events"][m]["p"]):
                    stats["events"]["extended_hypoglycemic_events"][m]["h"] = np.nan
                else:
                    stats["events"]["extended_hypoglycemic_events"][m]["h"] = 1 * (t.pvalue < alpha)
            else:
                if is_paired:
                    idxs = np.where(np.logical_and(~np.isnan(r1), ~np.isnan(r2)))[0]
                    if np.all(r1[idxs] - r2[idxs]) == 0:
                        stats["events"]["extended_hypoglycemic_events"][m]["h"] = np.nan
                        stats["events"]["extended_hypoglycemic_events"][m]["p"] = np.nan
                    else:
                        t = wilcoxon(r1, r2, nan_policy="omit")
                        stats["events"]["extended_hypoglycemic_events"][m]["p"] = t.pvalue
                        if np.isnan(stats["events"]["extended_hypoglycemic_events"][m]["p"]):
                            stats["events"]["extended_hypoglycemic_events"][m]["h"] = np.nan
                        else:
                            stats["events"]["extended_hypoglycemic_events"][m]["h"] = 1 * (t.pvalue < alpha)
                else:
                    t = mannwhitneyu(r1, r2)
                    stats["events"]["extended_hypoglycemic_events"][m]["p"] = t.pvalue
                    if np.isnan(stats["events"]["extended_hypoglycemic_events"][m]["p"]):
                        stats["events"]["extended_hypoglycemic_events"][m]["h"] = np.nan
                    else:
                        stats["events"]["extended_hypoglycemic_events"][m]["h"] = 1 * (t.pvalue < alpha)

        return results, stats



import numpy as np

from py_agata.input_validator import *

def time_in_target(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the target range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_target: float
        The time percentage spent in target range.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th_l = 70.
        th_h = 180.
    elif glycemic_target == 'pregnancy':
        th_l = 63.
        th_h = 140.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_range(data=data, th_l=th_l, th_h=th_h, include_th_l=False, include_th_h=False)


def time_in_tight_target(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the tight target range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_tight_target: float
        The time percentage spent in tight target range.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th_l = 70.
        th_h = 140.
    elif glycemic_target == 'pregnancy':
        th_l = 70.
        th_h = 140.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_range(data=data, th_l=th_l, th_h=th_h, include_th_l=False, include_th_h=False)


def time_in_hypoglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the hypoglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_hypoglycemia: float
        The time percentage spent in hypoglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th = 70.
    elif glycemic_target == 'pregnancy':
        th = 63.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_below_range(data=data, th=th, include_th=True)


def time_in_l1_hypoglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the l1 hypoglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_l1_hypoglycemia: float
        The time percentage spent in l1 hypoglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th_l = 54.
        th_h = 70.
    elif glycemic_target == 'pregnancy':
        th_l = 54.
        th_h = 63.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_range(data=data, th_l=th_l, th_h=th_h, include_th_l=False, include_th_h=True)


def time_in_l2_hypoglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the l2 hypoglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_l2_hypoglycemia: float
        The time percentage spent in l2 hypoglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th = 54.
    elif glycemic_target == 'pregnancy':
        th = 54.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_below_range(data=data, th=th, include_th=True)


def time_in_hyperglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the hyperglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_hyperglycemia: float
        The time percentage spent in hyperglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th = 180.
    elif glycemic_target == 'pregnancy':
        th = 140.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_above_range(data=data, th=th, include_th=True)


def time_in_l1_hyperglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the l1 hyperglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_l1_hyperglycemia: float
        The time percentage spent in l1 hyperglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th_l = 180.
        th_h = 250.
    elif glycemic_target == 'pregnancy':
        th_l = 140.
        th_h = 250.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_range(data=data, th_l=th_l, th_h=th_h, include_th_l=True, include_th_h=False)


def time_in_l2_hyperglycemia(data, glycemic_target='diabetes'):
    """
    Computes the time spent in the l2 hyperglycemic range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    glycemic_target: str, {'diabetes', 'pregnancy'}, optional, default: 'diabetes'
        A string defining the set of glycemic targets to use. The default
        value is `diabetes`. It can be {`diabetes`,`pregnancy`).

    Returns
    -------
    time_in_l2_hyperglycemia: float
        The time percentage spent in l2 hyperglycemia.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Set the threshold
    if glycemic_target == 'diabetes':
        th = 250.
    elif glycemic_target == 'pregnancy':
        th = 250.
    else:
        raise RuntimeError('`glycemic_target` can be `diabetes` or `pregnancy`.')

    # Return the result
    return time_in_given_above_range(data=data, th=th, include_th=True)


def time_in_given_range(data, th_l, th_h, include_th_l=False, include_th_h=False):
    """
    Computes the time spent between a given range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    th_l: float
        The low level threshold of the range of interest (in mg/dl).
    th_h: float
        The high level threshold of the range of interest (in mg/dl).
    include_th_l: bool, optional, default: False
        A flag indicating whether to include or not th_l in the range of interest.
    include_th_h: bool, optional, default: False
        A flag indicating whether to include or not th_h in the range of interest.

    Returns
    -------
    time_in_given_range: float
        The time percentage spent in the given range.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th_l)
    check_float_parameter(th_h)

    # Get non-nan values
    values = data.glucose.values[~np.isnan(data.glucose.values)]

    # Return nan if all values are nan
    if values.size == 0:
        return np.nan

    # Get low/high flags
    flags_l = values >= th_l if include_th_l else values > th_l
    flags_h = values <= th_h if include_th_h else values < th_h

    # Return the results
    return 100 * np.where(np.logical_and(flags_l, flags_h))[0].shape[0]/values.shape[0]


def time_in_given_above_range(data, th, include_th=False):
    """
    Computes the time spent above a given range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    th: float
        The threshold of the range of interest (in mg/dl).
    include_th: bool, optional, default: False
        A flag indicating whether to include or not th in the range of interest.

    Returns
    -------
    time_in_given_above_range: float
        The time percentage spent above the given range.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th)

    # Get non-nan values
    values = data.glucose.values[~np.isnan(data.glucose.values)]

    # Return nan if all values are nan
    if values.size == 0:
        return np.nan

    # Get flags
    flags = values >= th if include_th else values > th

    # Return the results
    return 100 * np.where(flags)[0].shape[0]/values.shape[0]


def time_in_given_below_range(data, th, include_th=False):
    """
    Computes the time spent below a given range (ignoring nan values).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe with a column `glucose` containing the glucose data
        to analyze (in mg/dl).
    th: float
        The threshold of the range of interest (in mg/dl).
    include_th: bool, optional, default: False
        A flag indicating whether to include or not th in the range of interest.

    Returns
    -------
    time_in_given_below_range: float
        The time percentage spent below the given range.

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and metrics for clinical
    trials: An international consensus statement", The Lancet Diabetes &
    Endocrinology, 2022, pp. 1-16. DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th)

    # Get non-nan values
    values = data.glucose.values[~np.isnan(data.glucose.values)]

    # Return nan if all values are nan
    if values.size == 0:
        return np.nan

    # Get flags
    flags = values <= th if include_th else values < th

    # Return the results
    return 100 * np.where(flags)[0].shape[0]/values.shape[0]



import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta
import os
from py_agata.variability import modd


def test_modd():
    """
    Unit test of modd function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    start_time = datetime(2000, 1, 1, 0, 10, 0)
    end_time = datetime(2000, 1, 1, 0, 0, 0) + timedelta(minutes=4000)
    time_interval = timedelta(minutes=5)
    time_range = pd.date_range(start_time, end_time, freq=time_interval)

    # Set a random seed for reproducibility
    np.random.seed(1)

    # Generate random glucose data
    glucose = np.random.randn(len(time_range)) * 70 + 140

    # Create a DataFrame
    d = {'t': time_range, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    #Tests
    assert np.isnan(modd(data)) == False
    assert np.round(modd(data) * 1000) / 1000 == 82.866

    # Set empty data
    t = np.arange(datetime(2000, 1, 1, 1, 0, 0), datetime(2000, 1, 1, 1, 55, 0), timedelta(minutes=5)).astype(
        datetime)
    glucose = np.zeros(shape=(t.shape[0],))
    glucose[0] = np.nan
    glucose[1:3] = [np.nan, np.nan]
    glucose[3] = np.nan
    glucose[4:6] = [np.nan, np.nan]
    glucose[6:8] = [np.nan, np.nan]
    glucose[8:10] = [np.nan, np.nan]
    glucose[10] = np.nan
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    # Tests
    assert np.isnan(modd(data))


import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.inspection import number_days_of_observation


def test_number_days_of_observation():
    """
    Unit test of number_days_of_observation function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0) + timedelta(minutes=425),
                  timedelta(minutes=5)).astype(datetime)
    glucose = np.ones(shape=(t.shape[0],))*120
    glucose[80] = np.nan
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    #Tests
    assert np.isnan(number_days_of_observation(data)) == False
    assert np.round(number_days_of_observation(data)*100)/100 == 0.29

    # Set empty data
    d = {'t': [], 'glucose': []}
    data = pd.DataFrame(data=d)

    # Tests
    assert np.isnan(number_days_of_observation(data))



import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.processing import impute_glucose


def test_impute_glucose():
    """
    Unit test of impute_glucose function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0)+timedelta(minutes=125), timedelta(minutes=5)).astype(
        datetime)
    glucose = np.ones(shape=(t.size,))*120
    glucose[1:3] = [np.nan, np.nan]
    glucose[9:20] = [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    glucose[21] = np.nan
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    #Tests
    results = impute_glucose(data, 15)

    assert results.glucose.values.size == data.glucose.values.size
    assert (results.glucose[1] == 120)
    assert (results.glucose[2] == 120)
    assert (results.glucose[21] == 120)
    assert np.all(np.isnan(data.glucose.values[9:20]))


import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.processing import detrend_glucose


def test_detrend_glucose():
    """
    Unit test of detrend_glucose function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0)+timedelta(minutes=100), timedelta(minutes=5)).astype(
        datetime)
    glucose = np.arange(100,200,5)*1.0
    glucose[0:3] = [np.nan, np.nan, np.nan]
    glucose[9:20] = [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    #Tests
    results = detrend_glucose(data)

    assert (results.glucose[3] == 100)
    assert (results.glucose[4] == 100)
    assert (results.glucose[5] == 100)
    assert (results.glucose[6] == 100)
    assert (results.glucose[7] == 100)
    assert (results.glucose[8] == 100)

    assert np.all(np.isnan(data.glucose.values[0:3]))
    assert np.all(np.isnan(data.glucose.values[9:20]))

    # Set test data
    d = {'t': t[0], 'glucose': [100]}
    data = pd.DataFrame(data=d)

    # Tests
    results = detrend_glucose(data)
    assert (results.glucose[0] == 100)


import pandas as pd
import numpy as np
import datetime
from datetime import datetime, timedelta

from py_agata.inspection import find_hypoglycemic_events


def test_find_hypoglycemic_events():
    """
    Unit test of find_hypoglycemic_events function.

    Parameters
    ----------
    None

    Returns
    -------
    None

    Raises
    ------
    None

    See Also
    --------
    None

    Examples
    --------
    None

    References
    ----------
    None
    """
    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0) + timedelta(minutes=425),
                  timedelta(minutes=5)).astype(datetime)
    glucose = np.ones(shape=(t.shape[0],))*120
    glucose[9:13] = np.ones(shape=(4,))*50
    glucose[29:60] = np.ones(shape=(31,))*50
    glucose[31:33] = [np.nan, np.nan]
    glucose[61:63] = np.ones(shape=(2,)) * 50
    glucose[69:72] = np.ones(shape=(3,)) * 50
    glucose[75:78] = np.ones(shape=(3,)) * 50
    glucose[79:82] = np.ones(shape=(3,)) * 50
    glucose[80] = np.nan

    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    # Hypoglycemic events generated:
    #   - 2000-01-01 @ 00:45 - duration 00:20 (normal hypoglycemic event)
    #   - 2000-01-01 @ 02:45 - duration 02:30 (extended hypoglycemic event)
    #   - 2000-01-01 @ 05:45 - duration 00:15 (normal hypoglycemic event)
    #   - 2000-01-01 @ 06:15 - duration 00:35 (normal hypoglycemic event)

    # 1: check results structure
    results = find_hypoglycemic_events(data)
    assert type(results) is dict
    assert results['time_start'].size == results['duration'].size
    assert results['time_end'].size == results['duration'].size

    # 2. check results (wth events)
    assert results['time_start'][0] == datetime(2000, 1, 1, 0, 45, 0)
    assert results['time_end'][0] == datetime(2000, 1, 1, 0, 45, 0) + timedelta(minutes=20)
    assert results['duration'][0] == 20
    assert results['time_start'][1] == datetime(2000, 1, 1, 2, 45, 0)
    assert results['time_end'][1] == datetime(2000, 1, 1, 2, 45, 0) + timedelta(minutes=150)
    assert results['duration'][1] == 150
    assert results['time_start'][2] == datetime(2000, 1, 1, 5, 45, 0)
    assert results['time_end'][2] == datetime(2000, 1, 1, 5, 45, 0) + timedelta(minutes=15)
    assert results['duration'][2] == 15
    assert results['time_start'][3] == datetime(2000, 1, 1, 6, 15, 0)
    assert results['time_end'][3] == datetime(2000, 1, 1, 6, 15, 0) + timedelta(minutes=35)
    assert results['duration'][3] == 35
    assert results['time_start'].size == results['duration'].size
    assert results['time_end'].size == results['duration'].size
    assert results['time_start'].size == 4
    assert results['mean_duration'] == 55
    assert results['events_per_week'] == 96

    # 3. check results (with custom threshold)
    results = find_hypoglycemic_events(data, th=45.)
    assert results['time_start'].size == results['duration'].size
    assert results['time_end'].size == results['duration'].size
    assert results['time_start'].size == 0
    assert np.isnan(results['mean_duration'])
    assert results['events_per_week'] == 0

    # 4. check results (without events)
    data.glucose = np.ones(shape=(data.glucose.values.size,))*120
    results = find_hypoglycemic_events(data)
    assert results['time_start'].size == results['duration'].size
    assert results['time_end'].size == results['duration'].size
    assert results['time_start'].size == 0
    assert np.isnan(results['mean_duration'])
    assert results['events_per_week'] == 0


    # Set empty data
    d = {'t': [], 'glucose': []}
    data = pd.DataFrame(data=d)

    # 5. Test with empty data
    results = find_hypoglycemic_events(data)
    assert results['time_start'].size == results['duration'].size
    assert results['time_end'].size == results['duration'].size
    assert results['time_start'].size == 0
    assert np.isnan(results['mean_duration'])
    assert np.isnan(results['events_per_week'])

    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0) + timedelta(minutes=435),
                  timedelta(minutes=5)).astype(datetime)
    glucose = np.ones(shape=(t.shape[0],)) * 120
    glucose[9:13] = np.ones(shape=(4,)) * 50
    glucose[29:60] = np.ones(shape=(31,)) * 50
    glucose[31:33] = [np.nan, np.nan]
    glucose[61:63] = np.ones(shape=(2,)) * 100
    glucose[69:72] = np.ones(shape=(3,)) * 100
    glucose[75:78] = np.ones(shape=(3,)) * 100
    glucose[79:82] = np.ones(shape=(3,)) * 100
    glucose[84:87] = np.ones(shape=(3,)) * 50
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    # 6. Test last entry
    results = find_hypoglycemic_events(data)
    assert results['time_start'][-1] == datetime(2000, 1, 1, 7, 0, 0)
    assert results['time_end'][-1] == datetime(2000, 1, 1, 7, 0, 0) + timedelta(minutes=15)
    assert results['duration'][-1] == 15

    # Set test data
    t = np.arange(datetime(2000, 1, 1, 0, 0, 0), datetime(2000, 1, 1, 0, 0, 0) + timedelta(minutes=440),
                  timedelta(minutes=5)).astype(datetime)
    glucose = np.ones(shape=(t.shape[0],)) * 120
    glucose[9:13] = np.ones(shape=(4,)) * 50
    glucose[29:60] = np.ones(shape=(31,)) * 50
    glucose[31:33] = [np.nan, np.nan]
    glucose[61:63] = np.ones(shape=(2,)) * 100
    glucose[69:72] = np.ones(shape=(3,)) * 100
    glucose[75:78] = np.ones(shape=(3,)) * 100
    glucose[79:82] = np.ones(shape=(3,)) * 100
    glucose[84:87] = np.ones(shape=(3,)) * 50
    glucose[87] = 100
    d = {'t': t, 'glucose': glucose}
    data = pd.DataFrame(data=d)

    # 6b. Test last entry
    results = find_hypoglycemic_events(data)
    assert results['time_start'][-1] == datetime(2000, 1, 1, 7, 0, 0)
    assert results['time_end'][-1] == datetime(2000, 1, 1, 7, 0, 0) + timedelta(minutes=15)
    assert results['duration'][-1] == 15


import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from copy import copy

from py_agata.input_validator import *

def find_nan_islands(data, th):
    """
    Locates nan sequences in vector `data`, and classifies them based on their length (longer or
    not than the specified threshold `th`).

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    th: int
        Threshold, expressed in number of samples, to distinguish between
        long and short nan sequences

    Returns
    -------
    short_nan: np.ndarray
        The indices of "short nan" sequences (i.e., sequences shorter than TH consecutive nan samples)
    long_nan: np.ndarray
        The indices of "long nan" sequences (i.e., sequences longer than or equal to TH consecutive nan samples)
    nan_start: np.ndarray
        Starts of each nan sequence
    nan_end: end of each nan sequence

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_int_parameter(th)

    # Get glucose data
    glucose = data.glucose.values

    # Locate nan sequences
    nan_ind = np.where(np.isnan(glucose))[0]

    if nan_ind.size == 0:
        return np.array([]), np.array([]), np.array([]), np.array([])
    elif nan_ind.size == 1:
        if th <= 1:
            return np.array([]), np.array([nan_ind[0]]), np.array([nan_ind[0]]), np.array([nan_ind[0]])
        else:
            return np.array([nan_ind[0]]), np.array([]), np.array([nan_ind[0]]), np.array([nan_ind[0]])
    else:

        short_nan = np.array([])
        long_nan = np.array([])
        nan_start = np.array([])
        nan_end = np.array([])

        start_tmp = nan_ind[0]
        for i in range(1, len(nan_ind)):
            if nan_ind[i] > nan_ind[i-1] + 1:
                nan_start = np.append(nan_start, start_tmp)
                nan_end = np.append(nan_end, nan_ind[i-1])

                if (nan_end[-1] - nan_start[-1] + 1) >= th:
                    long_nan = np.append(long_nan, np.arange(nan_start[-1], nan_end[-1] + 1))
                else:
                    short_nan = np.append(short_nan, np.arange(nan_start[-1], nan_end[-1] + 1))

                start_tmp = nan_ind[i]
        nan_start = np.append(nan_start, start_tmp)
        nan_end = np.append(nan_end, nan_ind[i])

        if (nan_end[-1] - nan_start[-1] + 1) >= th:
            long_nan = np.append(long_nan, np.arange(nan_start[-1], nan_end[-1] + 1))
        else:
            short_nan = np.append(short_nan, np.arange(nan_start[-1], nan_end[-1] + 1))

        return short_nan.astype(int), long_nan.astype(int), nan_start.astype(int), nan_end.astype(int)


def missing_glucose_percentage(data):
    """
    Computes the percentage of missing values in the given glucose trace.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data

    Returns
    -------
    missing_glucose_percentage: float
        The percentage of missing glucose values.

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    if data.glucose.values.size == 0:
        return np.nan

    return 100 * np.sum(np.isnan(data.glucose))/ data.glucose.values.size


def number_days_of_observation(data):
    """
    Computes the percentage of number of days of observation in the given glucose trace.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data

    Returns
    -------
    number_days_of_observation: float
        The number of days of observation.

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    if data.glucose.values.size == 0:
        return np.nan

    start_time = pd.to_datetime(data.t.values[0]).to_pydatetime()
    end_time = pd.to_datetime(data.t.values[-1]).to_pydatetime()
    return (end_time - start_time).seconds / (60 * 60 * 24)


def find_hypoglycemic_events(data, th=70.):
    """
    Finds the hypoglycemic events in a given glucose trace. The definition of hypoglycemic event can be found
    in Battellino et al. (event begins: at least consecutive 15 minutes < threshold mg/dl, event ends: at least
    15 consecutive minutes > threshold mg/dl)

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    th: float, optional, default : 70
        A number defining the threshold to be used to find hypoglycemia (in mg/dl)

    Returns
    -------
    hypoglycemic_events: dict
        A dictionary containing the information on the hypoglycemic events found by the function with fields:
        - time_start: np.ndarray
            The starting timestamps of each found hypoglycemic event
        - time_end: np.ndarray
            The ending timestamps of each found hypoglycemic event
        - duration: np.ndarray
            The duration of each found hypoglycemic event
        - mean_duration: float
            The mean duration of hypoglycemic events
        - number_per_week: float
            The number of hypoglycemic events per week

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and merics for
    clinical trials: An international consensus statement", The Lancet
    Diabetes & Endocrinology, 2022, pp. 1-16.
    DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th)

    hypoglycemic_events = dict()
    hypoglycemic_events['time_start'] = np.empty(shape=0, dtype=datetime)
    hypoglycemic_events['time_end'] = np.empty(shape=0, dtype=datetime)
    hypoglycemic_events['duration'] = np.empty(shape=0, dtype=float)
    hypoglycemic_events['mean_duration'] = np.nan
    hypoglycemic_events['events_per_week'] = np.nan
    if data.glucose.values.size == 0:
        return hypoglycemic_events

    k = 0 #hypoglycemic_event vector current index
    t0 = pd.to_datetime(data.t.values[0]).to_pydatetime()
    t1 = pd.to_datetime(data.t.values[1]).to_pydatetime()
    sample_time = (t1 - t0).seconds / 60

    n_samples = int(np.round(15/sample_time)) #number of consecutive samples required to define a valid event

    count = 0 #number of current found consecutive samples

    # state flag -> 0: no event, 1: found a valid hypo event and
    #               currently in hypo, -1: found a valid hypo event and
    #               currently not in hypo.
    flag = 0

    for t in range(data.glucose.values.size):

        if data.glucose.values[t] < th:

            # If it is a new event, reset count and set the hypothetical starting time to the current timestamp
            if count <= 0:
                count = 0
                temp_start_time = pd.to_datetime(data.t.values[t]).to_pydatetime()

            count = min([n_samples, count + 1]) #limit count to n_samples

            # if count touches the "nSamples sample goal" we found a new event
            if count == n_samples or flag == -1:
                flag = 1

        else:

            if flag == 0 and count > 0:
                count = 0

            elif flag == 1:
                count = n_samples
                flag = -1

            # countdown
            count = count - 1

        if count == 0 and flag == -1:
            hypoglycemic_events['time_start'] = np.append(hypoglycemic_events['time_start'], temp_start_time)
            hypoglycemic_events['duration'] = np.append(hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
            k += 1
            flag = 0

    # Manage the case in which the hypoglycemic event has been found (at
    # least n_samples < th) but the post hypoglycemic window has not finshed
    # yet.
    if count > 0 and flag == -1:
        hypoglycemic_events['time_start'] = np.append(hypoglycemic_events['time_start'], temp_start_time)
        hypoglycemic_events['duration'] = np.append(hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples - count - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
        k += 1

    if count == n_samples and flag == 1:
        hypoglycemic_events['time_start'] = np.append(hypoglycemic_events['time_start'], temp_start_time)
        hypoglycemic_events['duration'] = np.append(hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t]).to_pydatetime() - temp_start_time).seconds / 60 + sample_time)
        k = k + 1

    for k in range(hypoglycemic_events['time_start'].size):
        hypoglycemic_events['time_end'] = np.append(hypoglycemic_events['time_end'], hypoglycemic_events['time_start'][k] + timedelta(minutes=hypoglycemic_events['duration'][k]))

    if hypoglycemic_events['duration'].size == 0:
        hypoglycemic_events['mean_duration'] = np.nan
    else:
        hypoglycemic_events['mean_duration'] = np.mean(hypoglycemic_events['duration'])
    n_days = number_days_of_observation(data)
    hypoglycemic_events['events_per_week'] = hypoglycemic_events['time_start'].size / n_days * 7

    return hypoglycemic_events


def find_hyperglycemic_events(data, th=180.):
    """
    Finds the hyperglycemic events in a given glucose trace. The definition of hyperglycemic event can be found
    in Battellino et al. (event begins: at least consecutive 15 minutes > threshold mg/dl, event ends: at least
    15 consecutive minutes < threshold mg/dl)

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    th: float, optional, default : 180
        A number defining the threshold to be used to find hyperglycemia (in mg/dl)

    Returns
    -------
    hyperglycemic_events: dict
        A dictionary containing the information on the hyperglycemic events found by the function with fields:
        - time_start: np.ndarray
            The starting timestamps of each found hyperglycemic event
        - time_end: np.ndarray
            The ending timestamps of each found hyperglycemic event
        - duration: np.ndarray
            The duration of each found hyperglycemic event
        - mean_duration: float
            The mean duration of hyperglycemic events
        - number_per_week: float
            The number of hyperglycemic events per week

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and merics for
    clinical trials: An international consensus statement", The Lancet
    Diabetes & Endocrinology, 2022, pp. 1-16.
    DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th)

    hyperglycemic_events = dict()
    hyperglycemic_events['time_start'] = np.empty(shape=0, dtype=datetime)
    hyperglycemic_events['time_end'] = np.empty(shape=0, dtype=datetime)
    hyperglycemic_events['duration'] = np.empty(shape=0, dtype=float)
    hyperglycemic_events['mean_duration'] = np.nan
    hyperglycemic_events['events_per_week'] = np.nan
    if data.glucose.values.size == 0:
        return hyperglycemic_events

    k = 0 #hypoglycemic_event vector current index
    t0 = pd.to_datetime(data.t.values[0]).to_pydatetime()
    t1 = pd.to_datetime(data.t.values[1]).to_pydatetime()
    sample_time = (t1 - t0).seconds / 60

    n_samples = int(np.round(15/sample_time)) #number of consecutive samples required to define a valid event

    count = 0 #number of current found consecutive samples

    # state flag -> 0: no event, 1: found a valid hyper event and
    #               currently in hypo, -1: found a valid hyper event and
    #               currently not in hyper.
    flag = 0

    for t in range(data.glucose.values.size):

        if data.glucose.values[t] > th:

            # If it is a new event, reset count and set the hypothetical starting time to the current timestamp
            if count <= 0:
                count = 0
                temp_start_time = pd.to_datetime(data.t.values[t]).to_pydatetime()

            count = min([n_samples, count + 1]) #limit count to n_samples

            # if count touches the "nSamples sample goal" we found a new event
            if count == n_samples or flag == -1:
                flag = 1

        else:

            if flag == 0 and count > 0:
                count = 0

            elif flag == 1:
                count = n_samples
                flag = -1

            # countdown
            count = count - 1

        if count == 0 and flag == -1:
            hyperglycemic_events['time_start'] = np.append(hyperglycemic_events['time_start'], temp_start_time)
            hyperglycemic_events['duration'] = np.append(hyperglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
            k += 1
            flag = 0

    # Manage the case in which the hyperglycemic event has been found (at
    # least n_samples > th) but the post hyperglycemic window has not finshed
    # yet.
    if count > 0 and flag == -1:
        hyperglycemic_events['time_start'] = np.append(hyperglycemic_events['time_start'], temp_start_time)
        hyperglycemic_events['duration'] = np.append(hyperglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples - count - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
        k += 1

    if count == n_samples and flag == 1:
        hyperglycemic_events['time_start'] = np.append(hyperglycemic_events['time_start'], temp_start_time)
        hyperglycemic_events['duration'] = np.append(hyperglycemic_events['duration'], (pd.to_datetime(data.t.values[t]).to_pydatetime() - temp_start_time).seconds / 60 + sample_time)
        k = k + 1

    for k in range(hyperglycemic_events['time_start'].size):
        hyperglycemic_events['time_end'] = np.append(hyperglycemic_events['time_end'], hyperglycemic_events['time_start'][k] + timedelta(minutes=hyperglycemic_events['duration'][k]))

    if hyperglycemic_events['duration'].size == 0:
        hyperglycemic_events['mean_duration'] = np.nan
    else:
        hyperglycemic_events['mean_duration'] = np.mean(hyperglycemic_events['duration'])
    n_days = number_days_of_observation(data)
    hyperglycemic_events['events_per_week'] = hyperglycemic_events['time_start'].size / n_days * 7

    return hyperglycemic_events


def find_extended_hypoglycemic_events(data, th=54.):
    """
    Finds the extended hypoglycemic events in a given glucose trace. The definition of hypoglycemic event can be found
    in Battellino et al. (event begins: at least consecutive 120 minutes < threshold mg/dl, event ends: at least
    15 consecutive minutes > threshold mg/dl)

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    th: float, optional, default : 54
        A number defining the threshold to be used to find hypoglycemia (in mg/dl)

    Returns
    -------
    extended_hypoglycemic_events: dict
        A dictionary containing the information on the extended hypoglycemic events found by the function with fields:
        - time_start: np.ndarray
            The starting timestamps of each found extended hypoglycemic event
        - time_end: np.ndarray
            The ending timestamps of each found extended hypoglycemic event
        - duration: np.ndarray
            The duration of each found extended hypoglycemic event
        - mean_duration: float
            The mean duration of extended hypoglycemic events
        - number_per_week: float
            The number of extended hypoglycemic events per week

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and merics for
    clinical trials: An international consensus statement", The Lancet
    Diabetes & Endocrinology, 2022, pp. 1-16.
    DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_float_parameter(th)

    extended_hypoglycemic_events = dict()
    extended_hypoglycemic_events['time_start'] = np.empty(shape=0, dtype=datetime)
    extended_hypoglycemic_events['time_end'] = np.empty(shape=0, dtype=datetime)
    extended_hypoglycemic_events['duration'] = np.empty(shape=0, dtype=float)
    extended_hypoglycemic_events['mean_duration'] = np.nan
    extended_hypoglycemic_events['events_per_week'] = np.nan
    if data.glucose.values.size == 0:
        return extended_hypoglycemic_events

    k = 0 #hypoglycemic_event vector current index
    t0 = pd.to_datetime(data.t.values[0]).to_pydatetime()
    t1 = pd.to_datetime(data.t.values[1]).to_pydatetime()
    sample_time = (t1 - t0).seconds / 60

    n_samples_in = int(np.round(120/sample_time)) #number of consecutive samples required to define the start of a valid event
    n_samples_out = int(np.round(15/ sample_time))  # number of consecutive samples required to define the end of a valid event

    count = 0 #number of current found consecutive samples

    # state flag -> 0: no event, 1: found a valid hypo event and
    #               currently in hypo, -1: found a valid hypo event and
    #               currently not in hypo.
    flag = 0

    for t in range(data.glucose.values.size):

        if data.glucose.values[t] < th:

            # If it is a new event, reset count and set the hypothetical starting time to the current timestamp
            if count <= 0:
                count = 0
                temp_start_time = pd.to_datetime(data.t.values[t]).to_pydatetime()

            count = min([n_samples_in, count + 1]) #limit count to n_samples

            # if count touches the "nSamples sample goal" we found a new event
            if count == n_samples_in or flag == -1:
                flag = 1

        else:

            if flag == 0 and count > 0:
                count = 0

            elif flag == 1:
                count = n_samples_out
                flag = -1

            # countdown
            count = count - 1

        if count == 0 and flag == -1:
            extended_hypoglycemic_events['time_start'] = np.append(extended_hypoglycemic_events['time_start'], temp_start_time)
            extended_hypoglycemic_events['duration'] = np.append(extended_hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples_out - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
            k += 1
            flag = 0

    # Manage the case in which the hypoglycemic event has been found (at
    # least n_samples < th) but the post hypoglycemic window has not finshed
    # yet.
    if count > 0 and flag == -1:
        extended_hypoglycemic_events['time_start'] = np.append(extended_hypoglycemic_events['time_start'], temp_start_time)
        extended_hypoglycemic_events['duration'] = np.append(extended_hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t - (n_samples_out - count - 1)]).to_pydatetime() - temp_start_time).seconds / 60)
        k += 1

    if count == n_samples_in and flag == 1:
        extended_hypoglycemic_events['time_start'] = np.append(extended_hypoglycemic_events['time_start'], temp_start_time)
        extended_hypoglycemic_events['duration'] = np.append(extended_hypoglycemic_events['duration'], (pd.to_datetime(data.t.values[t]).to_pydatetime() - temp_start_time).seconds / 60 + sample_time)
        k = k + 1

    for k in range(extended_hypoglycemic_events['time_start'].size):
        extended_hypoglycemic_events['time_end'] = np.append(extended_hypoglycemic_events['time_end'], extended_hypoglycemic_events['time_start'][k] + timedelta(minutes=extended_hypoglycemic_events['duration'][k]))

    if extended_hypoglycemic_events['duration'].size == 0:
        extended_hypoglycemic_events['mean_duration'] = np.nan
    else:
        extended_hypoglycemic_events['mean_duration'] = np.mean(extended_hypoglycemic_events['duration'])
    n_days = number_days_of_observation(data)
    extended_hypoglycemic_events['events_per_week'] = extended_hypoglycemic_events['time_start'].size / n_days * 7

    return extended_hypoglycemic_events


def find_hypoglycemic_events_by_level(data, glycemic_target = 'diabetes'):
    """
    Finds the hypoglycemic events in a given glucose trace classifying them by level, i.e., hypo, level 1 hypo or level 2 hypo.
    The definition of hypoglycemic event can be found in Battellino et al.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    glycemic_target: str, optional, default : 'diabetes'
        A string defining the set of glycemic targets to use.

    Returns
    -------
    hypoglycemic_events: dict
        A dictionary containing the information on the hypoglycemic events found by the function divided by level with fields:
        - hypo: dict
            A dictionary containing the information on the hypoglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found hypoglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found hypoglycemic event
            - duration: np.ndarray
                The duration of each found hypoglycemic event
            - mean_duration: float
                The mean duration of hypoglycemic events
            - number_per_week: float
                The number of hypoglycemic events per week
        - l1: dict
            A dictionary containing the information on the l1 hypoglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found l1 hypoglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found l1 hypoglycemic event
            - duration: np.ndarray
                The duration of each found l1 hypoglycemic event
            - mean_duration: float
                The mean duration of l1 hypoglycemic events
            - number_per_week: float
                The number of l1 hypoglycemic events per week
        - l2: dict
            A dictionary containing the information on the l2 hypoglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found l2 hypoglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found l2 hypoglycemic event
            - duration: np.ndarray
                The duration of each found l2 hypoglycemic event
            - mean_duration: float
                The mean duration of l2 hypoglycemic events
            - number_per_week: float
                The number of l2 hypoglycemic events per week

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and merics for
    clinical trials: An international consensus statement", The Lancet
    Diabetes & Endocrinology, 2022, pp. 1-16.
    DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_str_parameter(glycemic_target)

    if glycemic_target == 'diabetes':
        th_l1 = 70.
        th_l2 = 54.
    else:
        th_l1 = 63.
        th_l2 = 54.

    # Get all hypoglycemic events
    all_hypo_events = find_hypoglycemic_events(data, th=th_l1)

    # Get L2 hypoglycemic events
    l2_hypo_events = find_hypoglycemic_events(data, th=th_l2)

    flag_l1_events = np.full((all_hypo_events['time_start'].size,), True)

    for h in range(l2_hypo_events['time_start'].size):

        distances = [(x - l2_hypo_events['time_start'][h]).total_seconds() for x in all_hypo_events['time_start']]
        idxs = np.where(np.array(distances) < 0)[0]
        if idxs.size > 0:
            flag_l1_events[idxs[-1]] = False

    hypoglycemic_events = dict()
    hypoglycemic_events['hypo'] = copy(all_hypo_events)
    hypoglycemic_events['l1'] = dict()
    hypoglycemic_events['l1']['time_start'] = copy(all_hypo_events['time_start'][flag_l1_events])
    hypoglycemic_events['l1']['time_end'] = copy(all_hypo_events['time_end'][flag_l1_events])
    hypoglycemic_events['l1']['duration'] = copy(all_hypo_events['duration'][flag_l1_events])
    hypoglycemic_events['l1']['mean_duration'] = np.mean(hypoglycemic_events['l1']['duration'])
    n_days = number_days_of_observation(data)
    hypoglycemic_events['l1']['events_per_week'] = hypoglycemic_events['l1']['time_start'].size / n_days * 7
    hypoglycemic_events['l2'] = copy(l2_hypo_events)

    return hypoglycemic_events


def find_hyperglycemic_events_by_level(data, glycemic_target='diabetes'):
    """
    Finds the hyperglycemic events in a given glucose trace classifying them by level, i.e., hyper, level 1 hyper or level 2 hyper.
    The definition of hyperglycemic event can be found in Battellino et al.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    glycemic_target: str, optional, default : 'diabetes'
        A string defining the set of glycemic targets to use.

    Returns
    -------
    hyperglycemic_events: dict
        A dictionary containing the information on the hyperglycemic events found by the function divided by level with fields:
        - hyper: dict
            A dictionary containing the information on the hyperglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found hyperglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found hyperglycemic event
            - duration: np.ndarray
                The duration of each found hyperglycemic event
            - mean_duration: float
                The mean duration of hyperglycemic events
            - number_per_week: float
                The number of hyperglycemic events per week
        - l1: dict
            A dictionary containing the information on the l1 hyperglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found l1 hyperglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found l1 hyperglycemic event
            - duration: np.ndarray
                The duration of each found l1 hyperglycemic event
            - mean_duration: float
                The mean duration of l1 hyperglycemic events
            - number_per_week: float
                The number of l1 hyperglycemic events per week
        - l2: dict
            A dictionary containing the information on the l2 hyperglycemic events found by the function with fields:
            - time_start: np.ndarray
                The starting timestamps of each found l2 hyperglycemic event
            - time_end: np.ndarray
                The ending timestamps of each found l2 hyperglycemic event
            - duration: np.ndarray
                The duration of each found l2 hyperglycemic event
            - mean_duration: float
                The mean duration of l2 hyperglycemic events
            - number_per_week: float
                The number of l2 hyperglycemic events per week

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    Battelino et al., "Continuous glucose monitoring and merics for
    clinical trials: An international consensus statement", The Lancet
    Diabetes & Endocrinology, 2022, pp. 1-16.
    DOI: https://doi.org/10.1016/S2213-8587(22)00319-9.
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_str_parameter(glycemic_target)

    if glycemic_target == 'diabetes':
        th_l1 = 180.
        th_l2 = 250.
    else:
        th_l1 = 140.
        th_l2 = 250.

    # Get all hyperglycemic events
    all_hyper_events = find_hyperglycemic_events(data, th=th_l1)

    # Get L2 hyperglycemic events
    l2_hyper_events = find_hyperglycemic_events(data, th=th_l2)

    flag_l1_events = np.full((all_hyper_events['time_start'].size,), True)

    for h in range(l2_hyper_events['time_start'].size):

        distances = [(x - l2_hyper_events['time_start'][h]).total_seconds() for x in all_hyper_events['time_start']]
        idxs = np.where(np.array(distances) < 0)[0]
        if idxs.size > 0:
            flag_l1_events[idxs[-1]] = False

    hyperglycemic_events = dict()
    hyperglycemic_events['hyper'] = copy(all_hyper_events)
    hyperglycemic_events['l1'] = dict()
    hyperglycemic_events['l1']['time_start'] = copy(all_hyper_events['time_start'][flag_l1_events])
    hyperglycemic_events['l1']['time_end'] = copy(all_hyper_events['time_end'][flag_l1_events])
    hyperglycemic_events['l1']['duration'] = copy(all_hyper_events['duration'][flag_l1_events])
    hyperglycemic_events['l1']['mean_duration'] = np.mean(hyperglycemic_events['l1']['duration'])
    n_days = number_days_of_observation(data)
    hyperglycemic_events['l1']['events_per_week'] = hyperglycemic_events['l1']['time_start'].size / n_days * 7
    hyperglycemic_events['l2'] = copy(l2_hyper_events)

    return hyperglycemic_events



import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from copy import copy

from scipy.interpolate import interp1d

from py_agata.inspection import find_nan_islands
from py_agata.input_validator import *

def detrend_glucose(data):
    """
    Detrends glucose data. To do that, the function computes the slope of the immaginary line that "links" the first
    and last glucose datapoints in the timeseries, then it "flatten" the entire timeseries according to that slope.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data

    Returns
    -------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the detrended glucose data

    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)

    # Compute the slope
    first_point = np.where(~np.isnan(data.glucose.values))[0]
    if first_point.size > 1:
        t0 = pd.to_datetime(data.t.values[0]).to_pydatetime()
        t1 = pd.to_datetime(data.t.values[1]).to_pydatetime()
        sample_time = (t1 - t0).total_seconds() / 60

        last_point = first_point[-1]
        first_point = first_point[0]
        m = (data.glucose.values[last_point] - data.glucose.values[first_point]) / ((last_point - first_point) * sample_time)

        # Detrend data
        data_detrented = copy(data)
        data_detrented.glucose[:] = data_detrented.glucose.values - m*np.arange(0,data_detrented.glucose.values.size)*sample_time
        return data_detrented
    else:
        return copy(data)

def impute_glucose(data, max_gap):
    """
    Imputes missing glucose data using linear interpolation. The function imputes only missing data gaps of maximum
    `max_gap` minutes. Gaps longer than `max_gap` minutes are ignored.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    max_gap: int
        An integer defining the maximum interpolable missing data gaps (in min)

    Returns
    -------
    data_imputed: pd.DataFrame
        Pandas dataframe, numpy array or float containing the imputed glucose data


    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)
    check_homogeneous_timegrid(data)
    check_int_parameter(max_gap)

    # Get the sample time
    t0 = pd.to_datetime(data.t.values[0]).to_pydatetime()
    t1 = pd.to_datetime(data.t.values[1]).to_pydatetime()
    sample_time = (t1 - t0).total_seconds() / 60

    # Find the interpolable gaps
    short_nan, long_nan, nan_start, nan_end = find_nan_islands(data, int(np.round(max_gap / sample_time)))

    # Impute data
    data_imputed = copy(data)
    idxs = np.arange(data_imputed.shape[0])
    f = interp1d(idxs[~np.isnan(data.glucose.values)], data.glucose.values[~np.isnan(data.glucose.values)], kind='linear')
    data_imputed.glucose.values[short_nan] = f(short_nan)

    return data_imputed

def retime_glucose(data, timestep):
    """
    Retimes the given `data` timetable to a  new timetable with homogeneous `timestep`. It puts nans where glucose
    datapoints are missing and it uses mean to solve conflicts (i.e., when two glucose datapoints have the same retimed timestamp.

    Parameters
    ----------
    data: pd.DataFrame
        Pandas dataframe, numpy array or float containing the glucose data
    timestep: int
        An integer defining the timestep to use in the new timetable.

    Returns
    -------
    data_retimed: pd.DataFrame
        Pandas dataframe, numpy array or float containing the retimed glucose data


    Raises
    ------
    None

    See Also
    --------
    None
Ï
    Examples
    --------
    None

    References
    ----------
    None
    """
    # Check input
    check_dataframe(data)
    check_data_columns(data)

    data_temp = copy(data)
    start_time = pd.to_datetime(data_temp.t.values[0]).to_pydatetime()
    start_time = start_time.replace(second=0)
    end_time = pd.to_datetime(data_temp.t.values[-1]).to_pydatetime()

    new_t = np.arange(start_time, end_time, timedelta(minutes=timestep)).astype(datetime)
    values = np.empty(new_t.size)
    values.fill(np.nan)

    dr = {'t': new_t, 'glucose': values, 'k' : values}
    data_retimed = pd.DataFrame(data=dr)

    data_temp = data_temp.drop(np.where(np.isnan(data_temp.glucose.values))[0]).reset_index().drop(columns='index')

    t_list = pd.to_datetime(data_retimed.t.values).to_pydatetime().tolist()

    for t in range(data_temp.shape[0]):

        # Find the nearest timestamp
        t_temp = pd.to_datetime(data_temp.t.values[t]).to_pydatetime()
        distances = [abs(x - t_temp).total_seconds() for x in t_list]
        idx_near = np.where(min(distances) == np.array(distances))[0][0]

        # Manage conflicts computing their average
        if np.isnan(data_retimed.glucose[idx_near]):
            data_retimed.glucose[idx_near] = data_temp.glucose[t]
            data_retimed.k[idx_near] = 1
        else:
            data_retimed.glucose[idx_near] += data_temp.glucose[t]
            data_retimed.k[idx_near] += 1

    # Compute the average and remove column 'k'
    data_retimed.glucose = np.divide(data_retimed.glucose, data_retimed.k)
    data_retimed = data_retimed.drop(columns='k')

    return data_retimed